/* A wrapper that runs mpirun with the right arguments */
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <stropts.h>
#include <termio.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include "sut.h"
#include "debug.h"
#include "nodes.h"

#define BUF_SIZE	1024
#define CMD_SIZE	4096

/*N machine_args
+	-m <machine_file> - 
		Specify a file with a list of destination machines.  It is incorrect
		to use this with the -M option.  This option must preceed all others.

.	-M <machine_list> -
		Specify a list of destination machines.  The entire list must be
		surrounded by quotes.  It is incorrect to use this with the -m option.
		This option must preceed all others.
 N*/

/*N machine_envs
-	PT_MACHINE_FILE -
		If PT_MACHINE_FILE is set, the command will attempt to open the file 
		whose name is stored in PT_MACHINE_FILE.  This will be used if neither 
		the -m or -M options are given.
  N*/

/*D

  ptexec - SUT execute

Description:

ptexec executes a given command on specified hosts

Command Line Arguments:

ptexec [-m <machine_file>|-M <machine_list>] <command> [<arguments>...]

.N machine_args

. <cmd> - The command to execute
- <arguments> - The arguments to the command

Environment Variables:

+ SH - If set, SH is used as the shell to execute the command instead of 'sh'
.N machine_envs

Examples:

To execute the program 'myprog' on the machines specified in the 
file "destination_machines" use

.vb
ptexec -m destination_machines myprog
.ve

To do the same thing, specifying the machines on the command line, use

.vb
ptexec -M "host1 host2 host3" myprog
.ve

Notes:

The command is executed by a shell, so a shell command can be given.

D*/

int main(int argc, char *argv[])
{
    int i, j, status, pid;
    int len;
    int all, optind;
    int num_machines;
    char *tmpfn;
    char buf[BUF_SIZE];
    char cmd[CMD_SIZE];
    char **mpirun_argv;
    char *sh;

    if (nodes_init(argc, argv, 0, &tmpfn, &num_machines, &all, &optind) < 0)
        return 1;

    /*****************************************/
    /* Get the arguments ready to run mpirun */
    /*****************************************/
    mpirun_argv = malloc((argc+8)*sizeof(char *));
    if (mpirun_argv == NULL){
        perror("Unable to allocate new argument vector");
        return 1;
    }

    j = 0;
    mpirun_argv[j++] = "mpirun";
    mpirun_argv[j++] = "-np";
    snprintf(buf, BUF_SIZE, "%d", num_machines);
    mpirun_argv[j++] = buf;

    if (!all) {
        mpirun_argv[j++] = "-m";
        mpirun_argv[j++] = tmpfn;
    }
    mpirun_argv[j++] = "-1";
    sh = getenv("SH");
    mpirun_argv[j++] = sh == NULL ? "sh" : sh;
    mpirun_argv[j++] = "-c";

    cmd[0] = '\0';
    len = 0;
    for (i = optind; i < argc; i++) {
        len += strlen(argv[i]);
        if ((len < 0) || (len > CMD_SIZE)) {
            fprintf(stderr, "Command too long\n");
            return 1;
        }
        strncat(cmd, argv[i], CMD_SIZE-len);

        len += sizeof(" ");
        if ((len < 0) || (len > CMD_SIZE)) {
            fprintf(stderr, "Command too long\n");
            return 1;
        }
        strncat(cmd, " ", CMD_SIZE-len);
    }

    mpirun_argv[j++] = cmd;

    mpirun_argv[j++] = NULL;

    /**************/
    /* Run mpirun */
    /**************/
    pid = fork();
    if (pid < 0){
        perror("Unable to fork");
        return 1;
    }
    if (pid == 0){
        pt_debug("Executing '%s'\n", cmd);
        execvp(mpirun_argv[0], mpirun_argv);
        return 1;
    }
    wait(&status);
    if (!WIFEXITED(status))
        fprintf(stderr, "%s exited abnormally\n", mpirun_argv[0]);

    pt_debug("%s exited with code %d\n", cmd, WEXITSTATUS(status));

    /****************************************/
    /* Remove the intermediate machine file */
    /****************************************/
    unlink(tmpfn);

    return 0;
}
