#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <string.h>
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include "sut.h"
#include "ptfps.h"
#include "debug.h"

static char translate_state(status_t s)
{
    switch (s) {
        case ST_RUNNABLE:
            return 'R';
        case ST_SLEEPING:
            return 'S';
        case ST_STOPPED:
            return 'T';
        case ST_ZOMBIE:
            return 'Z';
        case ST_UNKNOWN:
        default:
            return ' ';
    }
}

int print_process(process_t *p, char *buf, size_t width, char *format, 
        int print_header)
{
    char *tmp;
    char *fmt;
    time_t t;
    struct tm *now, *pstime;
    char stime_st[20];
    char time_st[20];
    int d, h, m, s;

    time(&t);
    now = localtime(&t);
    if (now == NULL) {
        snprintf(buf, width, "Unable to get local time\n");
        return 1;
    }

    t = get_stime(p);
    pstime = localtime(&t);
    if (pstime == NULL) {
        snprintf(buf, width, "Unable to get process start time\n");
        return 1;
    }

    if (now->tm_yday != pstime->tm_yday)
        fmt = "%b%d";
    else if (now->tm_year != pstime->tm_year)
        fmt = "%Y";
    else 
        fmt = "%H:%M";

    strftime(stime_st, 20, fmt, pstime);

    t = get_time(p);
    if (strcasecmp(format, "LONG") == 0) {
        snprintf(time_st, 20, "%ld:%02ld", t / 60, t % 60);
        tmp = get_wchan(p);
        if (*tmp != '-')
            tmp[6] = '\0';
        snprintf(buf, print_header ? 2 * width : width, 
                "%s" /* header */
                "%03lx " /* flags */
                "%5d " /* uid */
                "%5d " /* pid */
                "%5d " /* ppid */
                "%3d " /* pri */
                "%3d " /* nice */
                "%5ld " /* vsz */
                "%4ld " /* rss */
                "%-6s " /* wchan */
                "%c    " /* state */
                "%-8s " /* tty */
                "%6s " /* time */
                "%s ", /* command */
                print_header ? 
                "  F   UID   PID  PPID PRI  NI   VSZ  RSS WCHAN  STAT TTY        TIME COMMAND\n"
                : "", 
                get_flags(p), get_euid(p), get_pid(p), get_ppid(p), 
                get_pri(p), get_nice(p), get_vsz(p), get_rss(p),
                tmp, translate_state(get_state(p)), get_tty(p), 
                time_st, get_name(p));
        buf[strlen(buf)-1] = '\n';
        return 1;
    } else if (strcasecmp(format, "BSDu") == 0) {
        snprintf(time_st, 20, "%ld:%02ld", t / 60, t % 60);
        snprintf(buf, print_header ? 2 * width : width, 
                "%s" /* header */
                "%-8s " /* user */
                "%5d " /* pid */
                "%4.1f " /* %cpu */
                "%4.1f " /* %mem */
                "%5ld " /* vsz */
                "%4ld " /* rss */
                "%-8s " /* tty */
                "%c    " /* stat */
                "%5s " /* start */
                "%6s " /* time */
                "%s ", /* command */
                print_header ? 
                "USER       PID %CPU %MEM   VSZ  RSS TTY      STAT START   TIME COMMAND\n"
                : "",
                get_eusername(p), get_pid(p), get_pcpu(p), 
                get_pmem(p), get_vsz(p), get_rss(p), get_tty(p), 
                translate_state(get_state(p)), stime_st, time_st, get_name(p));
        buf[strlen(buf)-1] = '\n';
        return 1;
    } else if (strcasecmp(format, "FULL") == 0) {
        s = t % 60;
        m = (t / 60) % 60;
        h = (t / 3600) % 24;
        d = (t / 86400);

        if (d != 0)
            snprintf(time_st, 20, "%d-%02d:%02d:%02d", d, h, m, s);
        else
            snprintf(time_st, 20, "%02d:%02d:%02d", h, m, s);

        snprintf(buf, print_header ? 2 * width : width, 
                "%s" /* header */
                "%-8s " /* uid */
                "%5d " /* pid */
                "%5d  " /* ppid */
                "%c " /* C */
                "%5s " /* stime */
                "%-8s " /* tty */
                "%8s " /* time */
                "%s ", /* command */
                print_header ?
                "UID        PID  PPID  C STIME TTY          TIME CMD\n"
                : "",
                get_eusername(p), get_pid(p), get_ppid(p), '0',
                stime_st, get_tty(p), time_st, get_name(p));
        buf[strlen(buf)-1] = '\n';
        return 1;
    } else if (strcasecmp(format, "NORMAL") == 0) {
        s = t % 60;
        m = (t / 60) % 60;
        h = (t / 3600) % 24;
        d = (t / 86400);

        if (d != 0)
            snprintf(time_st, 20, "%d-%02d:%02d:%02d", d, h, m, s);
        else
            snprintf(time_st, 20, "%2d:%02d:%02d", h, m, s);

        snprintf(buf, print_header ? 2 * width : width, 
                "%s" /* header */
                "%5d " /* pid */
                "%-8s " /* tty */
                "%8s " /* time */
                "%s ", /* cmd */
                print_header ?
                "  PID TTY          TIME CMD\n"
                : "",
                get_pid(p), get_tty(p), time_st, get_cmd(p));
        buf[strlen(buf)-1] = '\n';
        return 1;
    } else {
        snprintf(buf, width, "Special formats not supported at this time ");
        buf[strlen(buf)-1] = '\n';
        return 1;
    }
}

int kill_process(process_t *p, int signo)
{
    return kill(get_pid(p), signo) < 0;
}

#define BUF_SIZE 1024
int exec_process(process_t *p, list_t *args)
{
    int c = 0;
    char buf[BUF_SIZE];
    char **argv;
    list_node_t *l;
    pid_t pid;
    int status;

    argv = malloc((args->size+1)*sizeof(char *));
    if (argv == NULL)
        return 0;

    for (l = args->head, c = 0; l != NULL; l = l->next, c++){
        if (strcmp(l->data, "{}") == 0){
            snprintf(buf, BUF_SIZE, "%d", get_pid(p));
            argv[c] = strdup(buf);
        } else {
            argv[c] = strdup(l->data);
        }

        if (argv[c] == NULL){
            for (; c >= 0; c--)
                free(argv[c]);
            return 0;
        }
    }
    argv[c] = NULL;

    pid = fork();
    if (pid < 0){
        for (; c >= 0; c--)
            free(argv[c]);
        return 0;
    }
    if (pid == 0){
        execvp(argv[0], argv);
        exit(1);
    }
    wait(&status);

    if (!WIFEXITED(status)){
        for (; c >= 0; c--)
            free(argv[c]);
        return 0;
    }

    for (; c >= 0; c--)
        free(argv[c]);
    return 1;
}
